# SPDX-FileCopyrightText: 2024 Blue Brain Project / EPFL
#
# SPDX-License-Identifier: AGPL-3.0-or-later

# Functions that implement random controls of a network.  There are two general types of models

# Probability: Random controls implemented by assigning a probability for each edge to be part of the control.
# Shuffle: Random controls implementing by shuffling edges according to certain rules.

# IMPORTS #
import logging
import numpy as np
import scipy.sparse as sp
import bigrandomgraphs as gm
LOG = logging.getLogger("connectome-analysis-randomization")
LOG.setLevel("INFO")
logging.basicConfig(format="%(asctime)s %(levelname)-8s %(message)s",
                    level=logging.INFO,
                    datefmt="%Y-%m-%d %H:%M:%S")

#######################################################
################# PROBABILITY MODELS  #################
#######################################################

def _dict_to_coo(adj, N):
    # Utility function to format dict into scipy.sparse.coo
    return sp.coo_matrix((np.ones(len(adj['row'])), (adj['row'], adj['col'])), shape=(N, N))


def run_ER(n, p, threads=8, seed=(None,None)):
    """Creates an Erdos Renyi digraph.

    Parameters
    ----------
    n : int
        Number of vertices
    p : float
        Edge probablity, must satisfy $0 \\le p \\le 1$
    threads : int
        Number of parallel threads to be used
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control

    Examples
    --------
    Setting n=3 and p=1 gives the complete digraph on 3 vertices:
    >>> connalysis.randomization.run_ER(3,1)
    {'row': [0, 0, 1, 1, 2, 2], 'col': [1, 2, 0, 2, 0, 1]}

    Raises
    ------
    AssertionError
        If p is not between 0 and 1

    """
    assert (p >= 0 and p <= 1), "p must be between 0 and 1"
    if seed[0]==None or seed[1]==None:
        adj = gm.ER(n,p,threads)
    else:
        adj = gm.ER(n,p,threads,seed[0],seed[1])
    return _dict_to_coo(adj,n)

def ER_model(adj, threads=8, seed=(None,None)):
    """Creates an Erdos Renyi digraph.

    Parameters
    ----------
    adj : sparse matrix or 2d-array
        Adjacency matrix
    threads : int
        Number of parallel threads to be used to generate model
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control


    Raises
    ------
    AssertionError
        If adj is not square

    """
    assert adj.shape[0] == adj.shape[1], "The matrix is not square"
    n = adj.shape[0]
    p = adj.astype(bool).sum()/((n)*(n-1))
    if isinstance(seed, int):
      seed=(seed,seed)
    return run_ER(n=n, p=p, threads=threads, seed=seed)



def run_SBM(n, probs, blocks, threads=8, seed=(None,None)):
    """Creates a random digraph using the stochastic block model.

    Parameters
    ----------
    n : int
        Number of vertices
    probs : numpy array of floats
        shape=(m,m) where m is the number of blocks.
        probs[i][j] is probability of an edge between block i and block j
    blocks : numpy array of ints
        shape=(n,). The i'th entry gives to which block vertex i belongs.
    threads : int
        Number of parallel threads to be used
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control

    Examples
    --------
    To create an SBM digraph on 4 vertices where the even to
    odd, or odd to even, vertices connect with high probablity (p=0.9)
    and the even to evens or odd to odds connect with low probability (p=0.1):
    >>> connalysis.randomization.run_SBM(4,np.array([[0.1,0.9],[0.9,0.1]]),np.array([0,1,0,1]))
    {'row': [0, 0, 1, 1, 1, 2, 2, 3, 3], 'col': [1, 3, 0, 2, 3, 1, 3, 0, 2]


    Raises
    ------
    TypeError
        If blocks contains non-integers

    References
    ----------
    [1] P.W. Holland, K. Laskey, S. Leinhardt,
    ["Stochastic Blockmodels: First Steps"](https://www.sciencedirect.com/science/article/pii/0378873383900217),
    Soc Networks, 5-2, pp. 109-137, 1982

    """

    if seed[0]==None or seed[1]==None:
        adj = gm.SBM(n, probs, blocks, threads)
    else:
        adj = gm.SBM(n, probs, blocks, threads, seed[0], seed[1])
    return _dict_to_coo(adj, n)


def run_DD2(n,a,b,xyz,threads=8, seed=(None,None)):
    r"""Creates a random digraph using the 2nd-order probability model.

    Parameters
    ----------
    n : int
        Number of vertices
    a : float
        Coefficient of probability function
    b : float
        Absolute value of power of exponent in probability function
    xyz : (n,3)-numpy array of floats
        Co-ordinates of vertices in $\mathbb{R}^3$
    threads : int
        Number of parallel threads to be used
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control

    See Also
    --------
    [conn_prob_2nd_order_model](modelling.md#src.connalysis.modelling.modelling.conn_prob_2nd_order_model) :
    The modelling function from which the parameters ``a`` and ``b``can be obtained.


    """
    if seed[0]==None or seed[1]==None:
        adj = gm.DD2(n,a,b,xyz,threads)
    else:
        adj = gm.DD2(n,a,b,xyz,threads,seed[0],seed[1])
    return _dict_to_coo(adj,n)

def run_DD2_model(adj, node_properties,
                  model_params_dd2=None, #an analysis that could be loaded from the pipeline
                  coord_names= ['x', 'y', 'z'],
                  threads=8, return_params=False, **config_dict):
    """Wrapper for fitting a model and generating a random control graph based on 2nd order distance dependence model.

    Parameters
    ----------
    adj : sparse matrix or 2d-array
        Adjacency matrix.
        If model_params_dd2 have already been computed, one can pass an empty matrix of the right size.
    node_properties : pandas.DataFrame
        DataFrame with information on the vertices of adj.
        It must have columns corresponding to the names of the coord_names to be used for distance computation (Default: ['x', 'y', 'z']).
    model_params_dd2 : pandas.DataFrame
        Optional input of pre-computed model parameters as data frame with rows corresponding to seeds of model estimation
        (single row if subsampling is not used) and columns 'exp_model_scale' and 'exp_model_exponent' for the model parameters.
        See modelling.conn_prob_2nd_order_model for details.
    coord_names : list
        Names of the coordinates (corresponding to columns in neuron properties table) based on which to compute Euclidean distance.
        Default: ['x', 'y', 'z']
    threads : int
        Number of parallel threads to be used.
    return_params : bool
        If True, returns model_params_dd2 in addition to the generated control.
    config_dict : dict
        Dictionary with 2nd order model building settings.
        See modelling.conn_prob_2nd_order_model for details.

    Returns
    -------
    coo_matrix
        Matrix of the generated control
    model_params_dd2
        pandas.DataFrame with model parameters (optional; if return_params is True)

    See Also
    --------
    [conn_prob_2nd_order_model](modelling.md#src.connalysis.modelling.modelling.conn_prob_2nd_order_model) :
    The modelling function from which model_params_dd2 can be obtained.

    """

    if model_params_dd2 is None:
        from .import modelling
        #TODO:  What to do if coord_names are also given in configdict and do not match coord_names?
        config_dict["coord_names"]=coord_names
        model_params_dd2 = modelling.conn_prob_2nd_order_model(adj, node_properties,**config_dict)
        LOG.warning("Fit parameters are used directly but should be checked by hand if the proper fit is obtained!")

    LOG.info("Run DD2 model with parameters: \n%s", model_params_dd2)

    n = adj.shape[0]
    a = model_params_dd2.mean(axis=0)['exp_model_scale']
    b = model_params_dd2.mean(axis=0)['exp_model_exponent']
    xyz = node_properties.loc[:,coord_names].to_numpy() #Make and assert that checks these columns exist!
    if len(coord_names)<3: #Extend by zeros if lower dimensional data was used to compute distance
        xyz=np.hstack([xyz,np.zeros((xyz.shape[0],3-xyz.shape[1]))])
    C=gm.DD2(n,a,b,xyz,threads)
    i=C['row']
    j=C['col']
    data=np.ones(len(i))
    if return_params==True:
        return sp.coo_matrix((data, (i, j)), [n,n]), model_params_dd2
    else:
        return sp.coo_matrix((data, (i, j)), [n,n])


def run_DD3(n,a1,b1,a2,b2,xyz,depths,threads=8, seed=(None,None)):
    r"""Creates a random digraph using the 2nd-order probability model.

    Parameters
    ----------
    n : int
        Number of vertices
    a1 : float
        Coefficient of probability function for negative depth
    b1 : float
        Absolute value of power of exponent in probability function for negative depth
    a2 : float
        Coefficient of probability function for positive depth
    b2 : float
        Absolute value of power of exponent in probability function for positive depth
    xyz : (n,3)-numpy array of floats
        Co-ordinates of vertices in $\mathbb{R}^3$
    threads : int
        Number of parallel threads to be used
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control


    See Also
    --------
    [conn_prob_3rd_order_model](modelling.md#src.connalysis.modelling.modelling.conn_prob_3rd_order_model) :
    The modelling function from which the parameters ``a1/a2`` and ``b1/b2``can be obtained.



    """
    if seed[0]==None or seed[1]==None:
        adj = gm.DD3(n,a1,b1,a2,b2,xyz,depths,threads)
    else:
        adj = gm.DD3(n,a1,b1,a2,b2,xyz,depths,threads,seed[0],seed[1])
    return _dict_to_coo(adj,n)


def run_DD2_block_pre(n, probs, blocks, xyz, threads=8, seed=(None,None)):
    r"""Creates a random digraph using a combination of the stochastic block model
       and the 2nd order distance dependent model. Such that the probability of an edge
       is given by the distance dependent equation, but the parameters of that equation
       vary depending on the block of the source of the edge.

    Parameters
    ----------
    n : int
        Number of vertices
    probs : numpy array of floats
        shape=(m,2) where m is the number of blocks.
        probs[i][0] is the coefficient of the distance dependent equation (value a) for source vertex i and
        probs[i][0] is the absolute value of power of exponent in the distance dependent equation (value b)
    blocks : numpy array of ints
        shape=(n,). The i'th entry is which block vertex i belongs to.
    xyz : (n,3)-numpy array of floats
        Co-ordinates of vertices in $\mathbb{R}^3$
    threads : int
        Number of parallel threads to be used
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control


    Raises
    ------
    TypeError
        If blocks contains non-integers

    See Also
    --------
    [run_SBM](randomization.md#src.connalysis.randomization.randomization.run_SBM):
    Function which runs the stochastic block model

    [run_DD2](randomization.md#src.connalysis.randomization.randomization.run_DD2) :
    Function which runs the 2nd distance dependent model

    [run_DD2_block](randomization.md#src.connalysis.randomization.randomization.run_DD2_block) :
    Similar function that also accounts for the block of the target vertex

    """

    if seed[0]==None or seed[1]==None:
        adj = gm.DD2_block_pre(n, probs, blocks, xyz, threads)
    else:
        adj = gm.DD2_block_pre(n, probs, blocks, xyz, threads, seed[0], seed[1])
    return _dict_to_coo(adj,n)


def run_DD2_block(n, probs, blocks, xyz, threads, seed=(None,None)):
    r"""Creates a random digraph using a combination of the stochastic block model
       and the 2nd order distance dependent model. Such that the probability of an edge
       is given by the distance dependent equation, but the parameters of that equation
       vary depending on the block of the source of the edge and block of the target.

    Parameters
    ----------
    n : int
        Number of vertices
    probs : numpy array of floats
        shape=(m,m,2) where m is the number of blocks. For source vertex i and target vertex j
        probs[i][j][0] is the coefficient of the distance dependent equation (value a) and
        probs[i][j][0] is the absolute value of power of exponent in the distance dependent equation (value b)
    blocks : numpy array of ints
        shape=(n,). The i'th entry is which block vertex i belongs to.
    xyz : (n,3)-numpy array of floats
        Co-ordinates of vertices in $\mathbb{R}^3$
    threads : int
        Number of parallel threads to be used
    seed : pair of ints
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control


    Raises
    ------
    TypeError
        If blocks contains non-integers

    See Also
    --------
    [run_DD2](randomization.md#src.connalysis.randomization.randomization.run_DD2) :
    Function which runs the 2nd distance dependent model

    [run_SBM](randomization.md#src.connalysis.randomization.randomization.run_SBM) :
    Function which runs the stochastic block model

    [run_DD2_block_pre](randomization.md#src.connalysis.randomization.randomization.run_DD2_block_pre) :
    Similar function that only accounts for the block of the source vertex

    """
    if seed[0]==None or seed[1]==None:
        adj = gm.DD2_block(n, probs, blocks, xyz, threads)
    else:
        adj = gm.DD2_block(n, probs, blocks, xyz, threads, seed[0], seed[1])
    return _dict_to_coo(adj,n)


#######################################################
################### SHUFFLE MODELS  ###################
#######################################################
def _seed_random_state(shuffler, seeder=np.random.seed):
    """Decorate a connectivity shuffler to seed it's random-state before execution.

    It is expected that the generator can be seeded calling `seeder(seed)`.
    """
    def seed_and_run_method(adj, neuron_properties=[], seed=None, **kwargs):
        """Reinitialize numpy random state using the value of seed among `kwargs`.
        doing nothing if no `seed` provided --- expecting an external initialization.
        """
        if seed is None:
            LOG.warning("No seed among keyword arguments")
        else:
            seeder(seed)

        return shuffler(adj, neuron_properties, **kwargs)

    return seed_and_run_method


####### SHUFFLE #######################
@_seed_random_state
def ER_shuffle(adj, neuron_properties=[], shuffle_type="sparse"):
    """Creates an Erdős Renyi digraph with exactly the same number of edges and weights
    as adj by shuffling the non-diagonla entries of adj.

    Parameters
    ----------
    adj : sparse matrix or 2d array
        Base digraph for which the ER control is built.
    shuffle_type : string
        If ``dense`` if first converts adj to a dense array and then shuffle.
        If ``sparse`` it keeps sparsity during the shuffle.
    seed : int
        Random seed to be used, if none is provided a seed is randomly selected

    Returns
    -------
    coo matrix
        Matrix of the generated control

    Raises
    ------
    AssertionError
        If adj is not not sparse and ``shuffle_type`` is set to sparse.  A sparse shuffle can only be done to a sparse matrix.

    """
    N = adj.shape[0]
    if shuffle_type == "dense":
        if sp.issparse(adj):
            adj = adj.toarray()
        above_diagonal = adj[np.triu_indices(N, k=1)]
        below_diagonal = adj[np.tril_indices(N, k=-1)]
        off_diagonal = np.concatenate([above_diagonal, below_diagonal])

        # Shuffle away from the diagonal
        np.random.shuffle(off_diagonal)
        adj[np.triu_indices(N, k=1)] = off_diagonal[0:N * (N - 1) // 2]
        adj[np.tril_indices(N, k=-1)] = off_diagonal[N * (N - 1) // 2:]
        # Return matrix
        return sp.coo_matrix(adj)
    elif shuffle_type == "sparse":
        # Keep matrix sparse
        assert sp.issparse(adj), "Matrix must be sparse for sparse shuffle"
        data = adj.data
        K = adj.nnz  # number of non zero entries
        T = N * (N - 1)  # number of entries away from the diagonal

        # Select K entries in np.arange(T) representing the indices of the array of
        # pairs of tuples of entries away from the diagonal
        # [(0,1), (0,2), (0,3), ...,(0,n),
        #  (1,0) (1,2) (1,3) ... (1,n),
        # ....
        #  (n,0) (n,1) (n,2) ... (n,n-1)]
        rnd_index = np.random.choice(T, size=K, replace=False)

        # Get corresponding row and column indices
        rows = rnd_index // (N - 1)  # row indices
        alpha = rnd_index % (N - 1)  # colum indices candidates shifted by 1
        cols = alpha + 1 - (alpha < rows).astype(int)  # column indices

        # Return matrix
        return sp.coo_matrix((data, (rows, cols)), shape=(N, N))


def configuration_model(M, seed = None):
    """Function to generate the configuration control model, obtained by
    shuffling the row and column of coo format independently, to create
    new coo matrix, then removing any multiple edges and loops.

    Parameters
    ----------
    adj : coo-matrix
        Adjacency matrix of a directed network.
    seed : int
        Random seed to be used

    Returns
    -------
    csr matrix
        Configuration model control of adj

    See Also
    --------
    [run_SBM](randomization.md#src.connalysis.randomization.randomization.run_SBM) :
    Function which runs the stochastic block model

    [run_DD2](randomization.md#src.connalysis.randomization.randomization.run_DD2) :
    Function which runs the 2nd distance dependent model
    """
    adj=M.copy().tocoo()
    generator = np.random.default_rng(seed)
    R = adj.row
    C = adj.col
    generator.shuffle(R)
    generator.shuffle(C)
    CM_matrix = sp.coo_matrix(([1]*len(R),(R,C)),shape=adj.shape).tocsr()
    CM_matrix.setdiag(0)
    CM_matrix.eliminate_zeros()
    return CM_matrix

def adjusted_ER(adj, seed=None):
    """Function to generate an Erdos  Renyi model with adjusted bidirectional connections.

    Parameters
    ----------
    adj :  csc_matrix
        Adjacency matrix of a directed network.
    seed : int
        Random seed to be used

    Returns
    -------
    csc_matrix
        Erdos Renyi shuffled control with additional reciprocal connections added at random
        to match the number of reciprocal connections of the original matrix.

    See Also
    --------
    [underlying_model](randomization.md#src.connalysis.randomization.randomization.underlying_model) :
    Function which returns a digraph with the same  underlying undirected graph
    and same number of reciprocal connections

    [bishuffled_model](randomization.md#src.connalysis.randomization.randomization.bishuffled_model) :
    Function which returns a digraph with shuffled reciprocal connections
    """
    from connalysis.network.topology import rc_submatrix
    from .rand_utils import adjust_bidirectional_connections
    generator = np.random.default_rng(seed)
    ER_matrix = ER_shuffle(adj, seed=seed).tocsc()
    bedges_to_add = int(rc_submatrix(adj).count_nonzero() -rc_submatrix(ER_matrix).count_nonzero())//2
    if bedges_to_add >= 0:
        return adjust_bidirectional_connections(ER_matrix, bedges_to_add, generator)
    else:
        LOG.info("Erdos-Renyi control has more reciprocal connections than original, so they are not adjusted.")
        return ER_matrix

def underlying_model(adj, seed: int=None):
    """Function to generate a digraph with the same  underlying undirected graph as adj
        and the same number of reciprocal connections

    Parameters
    ----------
    adj : csc_matrix
        Adjacency matrix of a directed network.
    seed : int
        Random seed to be used

    Returns
    -------
    csc_matrix
        Digraph with the same  underlying undirected graph as adj and the same number of reciprocal connections

    See Also
    --------
    [adjusted_ER](randomization.md#src.connalysis.randomization.randomization.adjusted_ER) :
    Function to generate an Erdos  Renyi model with adjusted bidirectional connections

    [bishuffled_model](randomization.md#src.connalysis.randomization.randomization.bishuffled_model) :
    Function which returns a digraph with shuffled reciprocal connections
    """
    from connalysis.network.topology import rc_submatrix
    from .rand_utils import  add_bidirectional_connections
    generator = np.random.default_rng(seed)
    target_bedges = int(rc_submatrix(adj).count_nonzero() / 2)
    ut_matrix = sp.triu(adj + adj.T)
    return add_bidirectional_connections(ut_matrix, target_bedges, generator)

def bishuffled_model(adj, seed = None):
    """Function to generate a digraph with shuffled reciprocal connections

    Parameters
    ----------
    adj : csc_matrix
        Adjacency matrix of a directed network.
    seed : int
        Random seed to be used

    Returns
    -------
    csc_matrix
        Digraph with shuffled reciprocal connections

    See Also
    --------
    [adjusted_ER](randomization.md#src.connalysis.randomization.randomization.adjusted_ER) :
    Function to generate an Erdos  Renyi model with adjusted bidirectional connections

    [underlying_model](randomization.md#src.connalysis.randomization.randomization.underlying_model) :
    Function which returns a digraph with the same  underlying undirected graph
    and same number of reciprocal connections
    """
    from connalysis.network.topology import rc_submatrix
    from .rand_utils import  add_bidirectional_connections, half_matrix
    generator = np.random.default_rng(seed)
    ut_bedges = sp.triu(rc_submatrix(adj))
    target_bedges = ut_bedges.count_nonzero()
    bedges1, bedges2 = half_matrix(ut_bedges, generator)
    return add_bidirectional_connections(adj - bedges1 - bedges2.T, target_bedges, generator)

#######################################################
################ GRAPH MODIFICATIONS  #################
#######################################################
def add_rc_connections_skeleta(adj,factors,dimensions=None, skeleta=None, threads=8, seed=0, return_skeleta=False):
    """Function to add reciprocal connections at random to adj on the skeleta of maximal simplices of adj

    Parameters
    ----------
    adj : sparse matrix
        Adjacency matrix of a directed network
    factors: int or dict
        Factor by which to multiply the reciprocal connections on the ``k``-skeleta of adj.  If factors is an int
        the same factor is used on all dimensions.  Otherwise, factors can be a dictionary with keys dimensions
        and values the factor by which to multiply the number of reciprocal connections on that dimensions.
    dimensions: array
        The dimensions at which to increase the number of reciprocal connections.  If ``None`` then all dimensions
        will be used
    skeleta: dict
        Dictionary with keys f'dimension_{dim}' for dim in dimensions and values binary sparse sub-matrices of adj
        on which reciprocal connections will be added.
    threads: int
        Number of threads on which to parallelize the skeleta computation if not pre-computed
    seed : int
        Random seed to be used to selecte edges that will become reciprocal

    Returns
    -------
    csc_matrix, dict
        Digraph with add reciprocal connections
        If return_skeleta=True it also returns the skeleta of maximal simplices of adj in the dimensions selected

    """
    adj=adj.tocsr()
    from connalysis.network.topology import rc_submatrix
    from .rand_utils import add_bidirectional_connections
    # Compute skeleton graphs if not precomputed
    if skeleta is None:
        from connalysis.network.topology import get_k_skeleta_graph
        max_simplices=True # Add option for all simplices?
        skeleta=get_k_skeleta_graph(adj, max_simplices=max_simplices, dimensions=dimensions, threads=threads)
    # Restrict to dimensions that contain simplices
    if dimensions is None:
        dimensions = np.array([int(key[10:]) for key in skeleta.keys()])
    else:
        dimensions = np.intersect1d(np.array([int(key[10:]) for key in skeleta.keys()]), dimensions)
    # Generate mapping between factors and dimensions or check the one provided
    if isinstance(factors,int):
        factors={dim:factors for dim in dimensions}
    else:
        assert isinstance(factors, dict), "factors must be int or dictionary"
        assert np.isin(dimensions, np.array(list(factors.keys()))).all(), "all dimensions must be a key in factors"

    # Add bidirectional connections
    generator=np.random.default_rng(seed)
    rc_add={dim:((factors[dim]-1)*(rc_submatrix(skeleta[f'dimension_{dim}']).sum()))//2 for dim in dimensions}
    print("Number of reciprocal connections added per dimension"); print(rc_add) # Remove or add verbose option
    M=adj.copy()
    for dim in dimensions:
        M+=add_bidirectional_connections(skeleta[f'dimension_{dim}'], rc_add[dim], generator).astype(bool)
    if return_skeleta:
        return M, skeleta
    else:
        return M

def add_rc_connections(adj,n_rc, seed=0):
    """Function to turn a fixed amount of unidirectional connections of adj into reciprocal connections.

    Parameters
    ----------
    adj : sparse matrix
        Adjacency matrix of a directed network
    n_rc : Number of reciprocal connections to be added
    seed : int
        Random seed to be used to selecte edges that will become reciprocal

    Returns
    -------
    matrix
        Digraph with n_rc more edges than adj, all of which form reciprocal connections
    """
    # TODO: Move this function from utils and change dependencies
    from .rand_utils import add_bidirectional_connections
    # Add bidirectional connections
    generator=np.random.default_rng(seed)
    return add_bidirectional_connections(adj, n_rc, generator).astype(bool)

def add_connections(adj,nc, seed=0,sparse_mode=True, max_iter=30):
    """Function add connections at random

    Parameters
    ----------
    adj : matrix
        Adjacency matrix of a directed network
    nc : Number of connections to be added
    seed : int
        Random seed to be used to selecte edges that will become reciprocal
    sparse_mode: bool
        If sparse_mode is ``True`` the matrix is generated iteratively restricting to a sparse format.
        If ``False`` adj is converted to dense and edges are added in a single step

    Returns
    -------
    bool matrix
        Digraph with nc more edges than adj
    """
    adj=adj.astype(bool)
    # Add bidirectional connections
    if sparse_mode:
        # TODO: Search for more efficient way to do this
        N=adj.shape[0]; E=adj.sum(); k=0 # Number nodes, target edges and iteration counter
        while adj.sum()< E +nc: #target number of edges
            if k>max_iter:
                print("More than max_it iterations tried, increase number of iterations or try sparse_mode =False")
                break
            den=(E+nc-adj.sum())/(N*N) #density of matrix added
            generator = np.random.default_rng(seed)
            A=sp.random(*adj.shape, density=den, format='csr', dtype = 'bool', random_state = generator)
            A.setdiag(0)
            adj=adj+A; k+=1
        adj.eliminate_zeros()
    else:
        if sp.issparse(adj): adj=adj.toarray()
        ul_ind = np.where(np.eye(*adj.shape) == 0) # non-diagonal indices
        zero_ind=np.where(adj[ul_ind]==0)
        generator = np.random.default_rng(seed)
        selection=zero_ind[0][generator.choice(zero_ind[0].shape[0], replace=False, size=nc)]
        adj[(ul_ind[0][selection],ul_ind[1][selection])]=1
    return adj



# SPDX-FileCopyrightText: 2024 Blue Brain Project / EPFL
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import networkx as nx
import numpy as np
import pandas as pd
from scipy.stats import hypergeom
from scipy.stats import binom
from scipy.linalg import eigvals
import scipy.sparse as sp
#TODO: MODIFY THE IMPORTS TO EXTERNAL IMPORTS

from .topology import node_degree, rc_submatrix, underlying_undirected_matrix

def closeness_connected_components(adj, directed=False, return_sum=True):
    """Compute the closeness of each connected component of more than 1 vertex
    
    Parameters
    ----------
    adj : array_like
        Adjacency matrix of the graph
    directed : bool
        If `True`, will be computed using strongly connected components and directed closeness.
    return_sum : bool
        If `True`, only one list will be returned, by summing over all the connected components.


    Returns
    -------
    array_like
        A single array( if `return_sum=True`) or a list of arrays of shape `n`, containting closeness of vertices in that component, or 0 if the vertex is not in the component. Closeness cannot be zero otherwise.

    """
    from sknetwork.ranking import Closeness
    from scipy.sparse.csgraph import connected_components

    matrix = sp.csr_matrix(adj)
    if directed:
        n_comp, comp = connected_components(matrix, directed=True, connection="strong")
    else:
        n_comp, comp = connected_components(matrix, directed=False)
        matrix = matrix + matrix.T  # we need to make the matrix symmetric

    closeness = Closeness()  # if matrix is not symmetric automatically use directed
    n = matrix.shape[0]
    all_c = []
    for i in range(n_comp):
        c = np.zeros(n)
        idx = np.where(comp == i)[0]
        sub_mat = matrix[np.ix_(idx, idx)]
        if sub_mat.getnnz() > 0:
            c[idx] = closeness.fit_predict(sub_mat)
            all_c.append(c)
    if return_sum:
        all_c = np.array(all_c)
        return np.sum(all_c, axis=0)
    else:
        return all_c

def connected_components(adj,directed=True, connection='weak', return_labels=False):
    """Returns a list of the size of the connected components of the graph
    
    Parameters
    ----------
    adj : array_like or sparse matrix
        Adjacency matrix of the graph
    directed : bool
        If `True`, will be compute connected components of the directed graph
    connection : str {'weak', 'strong'}
        If `weak`, it will compute the connected components of the underlying undirected graph. 
        If `strong`, it will compute strongly connected components of the directed graph.
    return_labels : bool
        If `True`, will return the labels of the connected components

    Returns
    -------
    array_like
        A list of the size of the connected components of the graph. 
        If return_labels == True, it also returns the list of labels of the connected components.
    """

    matrix = sp.csr_matrix(adj)

    comps=sp.csgraph.connected_components(matrix , directed=directed, 
                                          connection=connection, return_labels=True)
    comps_size=np.unique(comps[1], return_counts=True)[1]
    if return_labels:
        return comps_size, comps[1]
    else:
        return comps_size

def core_number(adj):
    """Returns the core number for each node.

        Parameters
        ----------
        adj : array_like or sparse matrix
            Adjacency matrix of the graph
        directed : bool
            If `True`, will be compute connected components of the directed graph
        connection : str {'weak', 'strong'}
            If `weak`, it will compute the connected components of the underlying undirected graph. 
            If `strong`, it will compute strongly connected components of the directed graph.
        return_labels : bool
            If `True`, will return the labels of the connected components

        Returns
        -------
        dict
            A dictionary with keys the indices of the nodes of adj and values their corresponding core number.

        Notes
        -----
        The k-core of a graph is the maximal subgraph that contains nodes of degree k or more in the induced subgraph.
        The core number of a node is the largest value k of a k-core containing that node.
        For directed graphs the total node degree is use, i.e., the sum of in-degree + out-degree.
    """
    # TODO: Implement directed (k,l) core and k-core of underlying undirected graph (very similar to this)
    # TODO: Filtered simplex counts with different weights on vertices (coreness, intersection)
    #  or on edges (strength of connection).    
    adj=sp.csr_matrix(adj)
    G = nx.from_scipy_sparse_array(adj)
    # Very inefficient (returns a dictionary!). TODO: Look for different implementation
    return nx.algorithms.core.core_number(G)


def density(adj, type="directed", skip_symmetry_check=False):
    """Returns the density of a matrix.

        Parameters
        ----------
        adj : array_like or sparse matrix
            Adjacency matrix of the graph
        type : str {'directed', 'undirected', 'reciprocal'}
            The type of the graph considered for the computation.
            If 'directed', the density as a directed graph is computed.
            If 'undirected', the density of the underlying undirected graph is computed.
            If 'reciprocal', the density of the underlying reciprocal graph is computed.
        skip_symmetry_check : bool
            If `True`, it will skip the check for symmetry of the matrix.

        Returns
        -------
        float
            The density of the graph.
    """
    assert adj.shape[0] == adj.shape[1], "The matrix must be square!"
    n=adj.shape[0]
    adj=sp.csr_matrix(adj).astype('bool')
    if type=="undirected":
        if not skip_symmetry_check:
            if (adj!=adj.T).nnz >0:
                print("The graph is directed. Taking the underlying undirected graph")
                adj=underlying_undirected_matrix(adj)
        return 2*adj.sum() / (n*(n-1))
    elif type=="directed":
        return adj.sum() / (n*(n-1))
    if type=="reciprocal":
        adj=rc_submatrix(adj)
        return 2*adj.sum() /  (n*(n-1))
    
###### TODO: Code in this block needs docstrings ######

def __make_expected_distribution_model_first_order__(adj, direction="efferent"):
    #TODO: Document, utility function used in COMMON NEIGHBOURS ANALYSIS
    if direction == "efferent":
        N = adj.sum(axis=1).mean()
        M = adj.shape[1]
    elif direction == "afferent":
        N = adj.sum(axis=0).mean()
        M = adj.shape[0]
    else:
        raise ValueError("Unknown direction: {0}".format(direction))
    expected = hypergeom(M, N, N)
    return expected


def distribution_number_of_common_neighbors(adj, neuron_properties=None, direction="efferent"):
    # TODO: Document, utility function used in COMMON NEIGHBOURS ANALYSIS
    adj = adj.tocsc().astype(int)
    if direction == "efferent":
        cn = adj * adj.transpose()
    elif direction == "afferent":
        cn = adj.transpose() * adj
    else:
        raise ValueError("Unknown direction: {0}".format(direction))
    cn = np.array(cn.todense())
    cn = cn[np.triu_indices_from(cn, 1)]
    bins = np.arange(0, cn.max() + 2)
    return pd.Series(np.histogram(cn, bins=bins)[0], index=bins[:-1])


def normalized_distribution_of_common_neighbors(adj, neuron_properties=None, direction="efferent"):
    # TODO: Document, utility function used in COMMON NEIGHBOURS ANALYSIS
    data = distribution_number_of_common_neighbors(adj, neuron_properties, direction=direction)
    expected = __make_expected_distribution_model_first_order__(adj, direction=direction)
    expected = expected.pmf(data.index) * data.sum()
    expected = pd.Series(expected, index=data.index)
    return (data - expected) / (data + expected)


def overexpression_of_common_neighbors(adj, neuron_properties=None, direction="efferent"):
    # TODO: Document, utility function used in COMMON NEIGHBOURS ANALYSIS
    data = distribution_number_of_common_neighbors(adj, neuron_properties, direction=direction)
    data_mean = (data.index.values * data.values).sum() / data.values.sum()
    ctrl = __make_expected_distribution_model_first_order__(adj, direction=direction)
    ctrl_mean = ctrl.mean()
    return (data_mean - ctrl_mean) / (data_mean + ctrl_mean)


def common_neighbor_weight_bias(adj, neuron_properties=None, direction="efferent"):
    # TODO: Document, utility function used in COMMON NEIGHBOURS ANALYSIS
    adj_bin = (adj.tocsc() > 0).astype(int)
    if direction == "efferent":
        cn = adj_bin * adj_bin.transpose()
    elif direction == "afferent":
        cn = adj_bin.transpose() * adj_bin

    return np.corrcoef(cn[adj > 0],
                       adj[adj > 0])[0, 1]


def common_neighbor_connectivity_bias(adj, neuron_properties=None, direction="efferent",
                                      cols_location=None, fit_log=False):
    # TODO: Document, utility function used in COMMON NEIGHBOURS ANALYSIS
    import statsmodels.formula.api as smf
    from patsy import ModelDesc
    from scipy.spatial import distance

    if adj.dtype == bool:
        adj = adj.astype(int)

    if direction == "efferent":
        cn = adj * adj.transpose()
    elif direction == "afferent":
        cn = adj.transpose() * adj

    input_dict = {"CN": cn.toarray().flatten(),
                  "Connected": adj.astype(bool).toarray().flatten()}

    if fit_log:
        input_dict["CN"] = np.log10(input_dict["CN"] + fit_log)
    formula_str = "CN ~ Connected"
    if cols_location is not None:
        formula_str = formula_str + " + Distance"
        dmat = distance.squareform(distance.pdist(neuron_properties[cols_location].values))
        input_dict["Distance"] = dmat.flatten()
    sm_model = ModelDesc.from_formula(formula_str)

    sm_result = smf.ols(sm_model, input_dict).fit()

    pval = sm_result.pvalues.get("Connected[T.True]", 1.0)
    mdl_intercept = sm_result.params["Intercept"]
    mdl_added = sm_result.params.get("Connected[T.True]", 0.0)
    mdl_distance = sm_result.params.get("Distance", 0.0)
    return pval, mdl_added / mdl_intercept, 100 * mdl_distance / mdl_intercept

###Computing connection probabilities

def connection_probability_within(adj, neuron_properties, max_dist=200, min_dist=0,
                                  columns=["ss_flat_x", "ss_flat_y"]):
    #TODO: Does neuron_properties here assumes the right data structure?
    if isinstance(neuron_properties, tuple):
        nrn_pre, nrn_post = neuron_properties
    else:
        nrn_pre = neuron_properties;
        nrn_post = neuron_properties
    D = distance.cdist(nrn_pre[columns], nrn_post[columns])
    mask = (D > min_dist) & (D <= max_dist)  # This way with min_dist=0 a neuron with itself is excluded
    return adj[mask].mean()

#TODO ADD CODE FROM CLUSTER TO COMPUTE PROBABILITY OF CONNECTION PER PATHWAY OR ANY OTHER PROPERTIE ON THE NEURON_PROPERTY.

def gini_curve(m, nrn, direction='efferent'):
    m = m.tocsc()
    if direction == 'afferent':
        degrees = np.array(m.sum(axis=0)).flatten()
    elif direction == 'efferent':
        degrees = np.array(m.sum(axis=1).flatten())
    else:
        raise Exception("Unknown value for argument direction: %s" % direction)
    cs = np.cumsum(np.flipud(sorted(degrees))).astype(float) / np.sum(degrees)
    return pd.Series(cs, index=np.linspace(0, 1, len(cs)))


def gini_coefficient(m, nrn, direction='efferent'):
    gc = gini_curve(m, nrn, direction=direction)
    A = gc.index.values
    B = gc.values
    return np.sum(np.diff(A) * (B[:-1] + B[1:]) / 2.0)


def _analytical_expected_gini_curve(m, direction='efferent'):
    if direction == 'afferent':
        N = m.shape[0] - 1
        C = m.shape[1] * N
    elif direction == 'efferent':
        N = m.shape[1] - 1
        C = m.shape[0] * N
    P = m.nnz / C
    # Only using degrees, not distribution of weigthts. TODO: Fix that
    x = np.arange(N, -1, -1)
    p = binom.pmf(x, N, P)
    A = np.cumsum(p) / p.sum()
    B = np.cumsum(p * x) / np.sum(p * x)
    return pd.Series(B, index=A)


def normalized_gini_coefficient(m, nrn, direction='efferent'):
    gc = gini_coefficient(m, nrn, direction=direction)
    ctrl = _analytical_expected_gini_curve(m, direction=direction)
    A = ctrl.index.values
    B = ctrl.values
    return 2 * (gc - np.sum(np.diff(A) * (B[:-1] + B[1:]) / 2.0))


def _bin_degrees(degrees):
    nbins = np.maximum(int(len(degrees) * 0.1), np.minimum(len(degrees), 30))
    mx = np.nanmax(degrees);
    mn = np.nanmin(degrees)
    bins = np.linspace(mn, mx + 1E-6 * (mx - mn), nbins + 1)
    degrees = np.digitize(degrees, bins=bins) - 1
    udegrees = np.arange(nbins)
    ret_x = 0.5 * (bins[:-1] + bins[1:])
    return ret_x, udegrees, degrees

###### TODO: End of block that needs docstrings ######

def rich_club_curve(m, direction='TOTAL'):
    """Compute the rich-club curve of the matrix m.

    Parameters
    ----------
    m : array_like or sparse matrix
        Adjacency matrix of the graph.
    direction : str {'OUT', 'IN', 'TOTAL'}
        'OUT' : Compute the rich-club curve for the out-degree.

        'IN' : Compute the rich-club curve for the in-degree.
        
        'TOTAL' : Compute the rich-club curve for the total degree i.e., in-degree + out-degree.
        
    Returns
    -------
    Pandas series
        With index the center of the binned degrees and values the rich club coefficient at that degree.
    Notes
    -----
    The rich-club coefficient is a measure of the tendency of high-degree nodes to form tightly interconnected communities.
    """
    m = sp.csc_matrix(m)
    if direction == 'TOTAL': direction = None 
    degrees = node_degree(m, direction=direction)
    if m.dtype == bool:
        udegrees = np.arange(1, degrees.max() + 1)
        ret_x = udegrees
    else:
        ret_x, udegrees, degrees = _bin_degrees(degrees)

    edge_counter = lambda i: (degrees >= i).sum() * ((degrees >= i).sum() - 1)  # number of pot. edges
    mat_counter = lambda i: m[np.ix_(degrees >= i, degrees >= i)].sum()  # number of actual edges
    ret = (np.array([mat_counter(i) for i in udegrees]).astype(float)
           / np.array([edge_counter(i) for i in udegrees]))
    return pd.Series(ret, index=pd.Index(ret_x, name="degree"))


def efficient_rich_club_curve(M, direction="TOTAL", pre_calculated_filtration=None, sparse_bin_set=False):
    """Fast computation of the rich-club curve of the matrix M with respect to a degree filtration and possibly any filtration chosen by the user.

    Parameters
    ----------
    M : array_like or sparse matrix
        Adjacency matrix of the graph.
    direction : str {'OUT', 'IN', 'TOTAL'}
        'OUT' : Compute the rich-club curve for the out-degree.

        'IN' : Compute the rich-club curve for the in-degree.
        
        'TOTAL' or None : Compute the rich-club curve for the total degree i.e., in-degree + out-degree.
    pre_calculated_filtration : pandas series 
        To provide user defined filtration values other than degree.  
        The index are the nodes of M and values the filtration values.
    sparse_bin_set : bool
        If False, all integer values between 0 and the maximum degree/filtration value will be used to generate the bins. 
        If True, unique values of the degrees/filtration will be used. 
        This is useful when the degrees/filtration values are sparse over the whole range.
    Returns
    -------
    Pandas series
        With index the center of the binned degrees/filtration and values the rich club coefficient at that degree.
    Notes
    -----
    The rich-club coefficient is a measure of the tendency of high-degree nodes (nodes with a high filtration value) to form tightly interconnected communities.
    """
    #TODO: Maybe expand the notes explaining this concept.
    if pre_calculated_filtration is not None:
        deg = pre_calculated_filtration
    else:
        if direction=="TOTAL": direction = None 
        deg=node_degree(M,direction=direction)
    M = M.tocoo()
    shape = M.shape
    M = pd.DataFrame.from_dict({"row": M.row, "col": M.col})
    
    

    if sparse_bin_set == False:
        degree_bins = np.arange(deg.max() + 2)
    elif sparse_bin_set == True:
        degree_bins = np.unique(np.append(deg, [0, deg.max() + 1]))
    degree_bins_rv = degree_bins[-2::-1]
    nrn_degree_distribution = np.histogram(deg.values, bins=degree_bins)[0]
    nrn_cum_degrees = np.cumsum(nrn_degree_distribution[-1::-1])
    nrn_cum_pairs = nrn_cum_degrees * (nrn_cum_degrees - 1)

    deg_arr = np.zeros(shape[0], dtype=int)
    deg_arr[deg.index.values] = deg.values

    deg = None

    con_degree = np.minimum(deg_arr[M["row"].values], deg_arr[M["col"].values])
    M = None
    con_degree = np.histogram(con_degree, bins=degree_bins)[0]

    cum_degrees = np.cumsum(con_degree[-1::-1])

    return pd.Series(cum_degrees / nrn_cum_pairs, degree_bins_rv)[::-1]

###### TODO: This block needs docstrings ######

def _analytical_expected_rich_club_curve(m, direction='efferent'):
    assert m.dtype == bool, "This function only works for binary matrices at the moment"
    indegree = np.array(m.sum(axis=0))[0]
    outdegree = np.array(m.sum(axis=1))[:, 0]

    if direction == 'afferent':
        degrees = indegree
    elif direction == 'efferent':
        degrees = outdegree
    else:
        raise Exception("Unknown value for argument direction: %s" % direction)

    udegrees = np.arange(1, degrees.max() + 1)
    edge_counter = lambda i: (degrees >= i).sum() * ((degrees >= i).sum() - 1)
    res_mn = []
    res_sd = []
    for deg in udegrees:
        valid = np.nonzero(degrees >= deg)[0]
        i_v = indegree[valid]
        i_sum_all = indegree.sum() - i_v
        i_sum_s = i_v.sum() - i_v
        o_v = outdegree[valid]
        S = np.array([hypergeom.stats(_ia, _is, o)
                      for _ia, _is, o in zip(i_sum_all, i_sum_s, o_v)])
        res_mn.append(np.sum(S[:, 0]) / edge_counter(deg))
        res_sd.append(np.sqrt(S[:, 1].sum()) / edge_counter(deg))  # Sum the variances, but divide the std
    df = pd.DataFrame.from_dict({"mean": np.array(res_mn),
                                 "std": np.array(res_sd)})
    df.index = udegrees
    return df


def generate_degree_based_control(M, direction="efferent"):
    # A shuffled version of a connectivity matrix that aims to preserve degree distributions.
    # If direction = "efferent", then the out-degree is exactly preserved, while the in-degree is
    # approximately preseved. Otherwise it's the other way around.
    if direction == "efferent":
        M = M.tocsr()
        idxx = np.arange(M.shape[1])
        p_out = np.array(M.mean(axis=0))[0]
    elif direction == "afferent":
        M = M.tocsc()
        idxx = np.arange(M.shape[0])
        p_out = np.array(M.mean(axis=1))[:, 0]
    else:
        raise ValueError()

    for col in range(M.shape[1]):
        p = p_out.copy()
        p[col] = 0.0
        p = p / p.sum()
        a = M.indptr[col]
        b = M.indptr[col + 1]
        M.indices[a:b] = np.random.choice(idxx, b - a, p=p, replace=False)
    return M


def _randomized_control_rich_club_curve(m, direction='efferent', n=10):
    res = []
    for _ in range(n):
        m_shuf = generate_degree_based_control(m, direction=direction)
        res.append(efficient_rich_club_curve(m_shuf))
    res = pd.concat(res, axis=1)
    #TODO: Something is wrong here. rr is not defined. Should it be res?
    #      But changing rr to res causing 
    df = pd.DataFrame.from_dict(
        {
            "mean": np.nanmean(res, axis=1),
            "std": np.nanstd(res, axis=1)
        }
    )
    df.index = res.index
    return df


def normalized_rich_club_curve(m, direction='efferent', normalize='std',
                               normalize_with="shuffled", **kwargs):
    assert m.dtype == bool, "This function only works for binary matrices at the moment"
    data = efficient_rich_club_curve(m, direction=direction)
    A = data.index.values
    B = data.values
    if normalize_with == "analytical":
        ctrl = _analytical_expected_rich_club_curve(m, direction=direction)
    elif normalize_with == "shuffled":
        ctrl = _randomized_control_rich_club_curve(m, direction=direction)
    Ar = ctrl.index.values
    mn_r = ctrl["mean"].values
    sd_r = ctrl["std"].values

    if normalize == 'mean':
        return pd.Series(B[:len(mn_r)] / mn_r, index=A[:len(mn_r)])
    elif normalize == 'std':
        return pd.Series((B[:len(mn_r)] - mn_r) / sd_r, index=A[:len(mn_r)])
    else:
        raise Exception("Unknown normalization: %s" % normalize)


def rich_club_coefficient(m, **kwargs):
    Bn = normalized_rich_club_curve(m, normalize='std', **kwargs).values
    return np.nanmean(Bn)

###### TODO: End of block that needs docstrings ######

def ccc(matrix):
    """Computes the classical clustering coefficient of a directed graph

    Parameters
    ----------
    matrix : numpy array
        the adjaceny matrix of an unweighted graph

    Returns
    -------
    Series
        The index is the node index in ``matrix`` and the value is the clustering coefficient of the node. 

    References
    ----------
    The formula is taken from the following paper.

    [1] G. Fagiolo, "Clustering in complex directed networks", 2006;
            [DOI: 10.1103/PhysRevE.76.026107]().

    [2] Conceição, Pedro, et al. "An application of neighbourhoods in digraphs to the classification of binary dynamics.", 2022
            [DOI: 10.1162/netn_a_00228](). 

    """
    # We only analyze the udnerlying connectivity not the weights
    matrix=matrix.astype(bool).astype(int)
    # Numerator 
    if sp.issparse(matrix):
        dense_matrix = matrix + matrix.transpose()
        numerator = 0.5*np.diag((dense_matrix ** 3).toarray())
    else:
        numerator = 0.5*np.diag(np.linalg.matrix_power(matrix + np.transpose(matrix), 3))
    # Denominator
    deg = node_degree(matrix) 
    denominator = (deg*(deg-1)-2*rc_submatrix(matrix).toarray().sum(axis=0))
    return numerator/denominator

def largest_strongly_connected_component(adjacency_matrix):
    """Computes the largest strongly connected component of the graph with adjacency matrix adjacency_matrix,
        and returns the adjacency matrix of said component

    Parameters
    ----------
    adjacency_matrix : numpy array
        the adjaceny matrix of the DiGraph as a numpy array


    Returns
    -------
    numpy array
        The adjacency matrix of the largest strongly connected component

    """
    if sp.issparse(adjacency_matrix):
        adjacency_matrix = adjacency_matrix.toarray()
    current_tribe_nx = np_to_nx(adjacency_matrix)
    largest_comp = max(nx.strongly_connected_components(current_tribe_nx), key=len)
    current_tribe_strong_nx = current_tribe_nx.subgraph(largest_comp)
    current_tribe_strong = nx_to_np(current_tribe_strong_nx)
    return current_tribe_strong


def nx_to_np(directed_graph):
    """Converts networkx digraph to numpy array of the adjacency matrix 

    Parameters
    ----------
    directed_graph : networkx DiGraph
        a directed graph

    Returns
    -------
    numpy array
        the adjaceny matrix of the DiGraph as a numpy array
    """
    return nx.to_numpy_array(directed_graph,dtype=int)


def np_to_nx(adjacency_matrix):
    """Converts numpy array of an adjacency matrix to a networkx digraph

    Parameters
    ----------
    adjacency_matrix : numpy array
        the adjaceny matrix of the DiGraph as a numpy array


    Returns
    -------
    networkx DiGraph
            a directed graph

    """
    return nx.from_numpy_array(adjacency_matrix,create_using=nx.DiGraph)


# #*************************************************************************#
# #Code taken from TriDy
# #TODO: NEEDS CHECKING AND DOCUMENTATION

# ##
# ## HELPER FUNCTIONS (SPECTRAL)
# ##

# def spectral_gap(matrix, thresh=10, param='low'):
# #  In: matrix
# # Out: float
#     current_spectrum = spectrum_make(matrix)
#     current_spectrum = spectrum_trim_and_sort(current_spectrum, threshold_decimal=thresh)
#     return spectrum_param(current_spectrum, parameter=param)



# def spectrum_make(matrix):
# #  In: matrix
# # Out: list of complex floats
#     assert np.any(matrix) , 'Error (eigenvalues): matrix is empty'
#     eigenvalues = eigvals(matrix)
#     return eigenvalues



# def spectrum_trim_and_sort(spectrum, modulus=True, threshold_decimal=10):
# #  In: list of complex floats
# # Out: list of unique (real or complex) floats, sorted by modulus
#     if modulus:
#         return np.sort(np.unique(abs(spectrum).round(decimals=threshold_decimal)))
#     else:
#         return np.sort(np.unique(spectrum.round(decimals=threshold_decimal)))



# def spectrum_param(spectrum, parameter):
# #  In: list of complex floats
# # Out: float
#     assert len(spectrum) != 0 , 'Error (eigenvalues): no eigenvalues (spectrum is empty)'
#     if parameter == 'low':
#         if spectrum[0]:
#             return spectrum[0]
#         else:
#             assert len(spectrum) > 1 , 'Error (low spectral gap): spectrum has only zeros, cannot return nonzero eigval'
#             return spectrum[1]
#     elif parameter == 'high':
#         assert len(spectrum) > 1 , 'Error (high spectral gap): spectrum has one eigval, cannot return difference of top two'
#         return spectrum[-1]-spectrum[-2]
#     elif parameter == 'radius':
#         return spectrum[-1]


# ##
# ## NONSPECTRAL PARAMETER FUNCTIONS
# ##




# # degree type parameters

# # tribe size

# def tribe_size(matrix):
#     return len(matrix)

# # reciprocal connections

# def reciprocal_connections(matrix, chief_only=False):
#     #TODO: MERGE THIS WITH RC FUNCTION
#     if chief_only:
#         rc_count = np.count_nonzero(np.multiply(matrix[0],np.transpose(matrix)[0]))
#     else:
#         rc_count = np.count_nonzero(np.multiply(matrix,np.transpose(matrix)))//2
#     return rc_count


# ##
# ## SPECTRAL PARAMETER FUNCTIONS
# ##


# # adjacency spectrum

# def asg(matrix, gap='high'):
#     return spectral_gap(matrix, param=gap)


# # transition probability spectrum

# def tpsg(matrix, in_deg=False, gap='high'):
# #  in: tribe matrix
# # out: float
#     current_matrix = tps_matrix(matrix, in_deg=in_deg)
#     return spectral_gap(current_matrix, param=gap)

# def tps_matrix(matrix, in_deg=False):
# #  in: tribe matrix
# # out: transition probability matrix
#     current_size = len(matrix)
#     if in_deg:
#         degree_vector = node_degree(matrix, direction='IN').values
#     else:
#         degree_vector = node_degree(matrix, direction='OUT').values
#     inverted_degree_vector = [0 if not d else 1/d for d in degree_vector]
#     return np.matmul(np.diagflat(inverted_degree_vector),matrix)


# # chung laplacian spectrum
# # source 1: Laplacians and the Cheeger inequality for directed graph (Fan Chung, 2005)
# # source 2: https://networkx.org/documentation/stable/reference/generated/networkx.linalg.laplacianmatrix.directed_laplacian_matrix.html

# def clsg(matrix, is_strongly_conn=False, gap='low'):
# #  in: tribe matrix
# # out: float
#     chung_laplacian_matrix = cls_matrix_fromadjacency(matrix, is_strongly_conn=is_strongly_conn)
#     return spectral_gap(chung_laplacian_matrix, param=gap)

# def cls_matrix_fromadjacency(matrix, is_strongly_conn=False):
# #  in: numpy array
# # out: numpy array
#     matrix_nx = np_to_nx(matrix)
#     return cls_matrix_fromdigraph(matrix_nx, matrix=matrix, matrix_given=True, is_strongly_conn=is_strongly_conn)

# def cls_matrix_fromdigraph(digraph, matrix=np.array([]), matrix_given=False, is_strongly_conn=False):
# #  in: networkx digraph
# # out: numpy array
#     digraph_sc = digraph
#     matrix_sc = matrix
#     # Make sure is strongly connected
#     if not is_strongly_conn:
#         largest_comp = max(nx.strongly_connected_components(digraph), key=len)
#         digraph_sc = digraph.subgraph(largest_comp)
#         matrix_sc = nx_to_np(digraph_sc)
#     elif not matrix_given:
#         matrix_sc = nx_to_np(digraph_sc)
#     # Degeneracy: scc has size 1
#     if not np.any(matrix_sc):
#         return np.array([[0]])
#     # Degeneracy: scc has size 2
#     elif np.array_equal(matrix_sc,np.array([[0,1],[1,0]],dtype=int)):
#         return np.array([[1,-0.5],[-0.5,1]])
#     # No degeneracy
#     else:
#         return nx.directed_laplacian_matrix(digraph_sc)


# # bauer laplacian spectrum
# # source: Normalized graph Laplacians for directed graphs (Frank Bauer, 2012)

# def blsg(matrix, reverse_flow=False, gap='high'):
# #  in: tribe matrix
# # out: float
#     bauer_laplacian_matrix = bls_matrix(matrix, reverse_flow=reverse_flow)
#     return spectral_gap(bauer_laplacian_matrix, param=gap)

# def bls_matrix(matrix, reverse_flow=False):
# #  in: tribe matrix
# # out: bauer laplacian matrix
#     #non_quasi_isolated = [i for i in range(len(matrix)) if matrix[i].any()]
#     #matrix_D = np.diagflat([np.count_nonzero(matrix[nqi]) for nqi in non_quasi_isolated])
#     #matrix_W = np.diagflat([np.count_nonzero(np.transpose(matrix)[nqi]) for nqi in non_quasi_isolated])
#     #return np.subtract(np.eye(len(non_quasi_isolated),dtype=int),np.matmul(inv(matrix_D),matrix_W))
#     current_size = len(matrix)
#     return np.subtract(np.eye(current_size,dtype='float64'),tps_matrix(matrix, in_deg=(not reverse_flow)))



# ##
# ## TO BE DELETED ONCE EVERYTHING WORKS
# ##


# # # degree
# # def degree(chief_index, matrix, vertex_index=0):
# #     current_tribe = tribe(chief_index, matrix)
# #     return degree_adjacency(current_tribe, vertex_index=vertex_index)

# # def degree_adjacency(matrix, vertex_index=0):
# #     return in_degree_adjacency(matrix, vertex_index=vertex_index)+out_degree_adjacency(matrix, vertex_index=vertex_index)


# # # in-degree
# # def in_degree(chief_index, matrix, vertex_index=0):
# #     current_tribe = tribe(chief_index, matrix)
# #     return in_degree_adjacency(current_tribe, vertex_index=vertex_index)

# # def in_degree_adjacency(matrix, vertex_index=0):
# #     return np.count_nonzero(np.transpose(matrix)[vertex_index])


# # # out-degree
# # def out_degree(chief_index, matrix, vertex_index=0):
# #     current_tribe = tribe(chief_index, matrix)
# #     return out_degree_adjacency(current_tribe, vertex_index=vertex_index)



# # def out_degree_adjacency(matrix, vertex_index=0):
# #     return np.count_nonzero(matrix[vertex_index])


# SPDX-FileCopyrightText: 2024 Blue Brain Project / EPFL
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Functions to average (functional or structural) metrics across simplices or neighborhoods.
Author(s): Daniela Egas Santander,
Last update: 11.2023
"""

import numpy as np
import pandas as pd
from scipy import stats
from tqdm import tqdm
import operator

from .local import neighborhood_indices


def node_stats_per_position_single(simplex_list, values, with_multiplicity=True):
    """ Get mean, standard deviation and standard error of the mean averaged across simplex lists and filtered per position
    Parameters
    ----------
    simplex_list : 2d-array
        Array of dimension (no. of simplices, dimension) listing simplices to be considered.
        Each row corresponds to a list of nodes on a simplex indexed by the order of the nodes in an NxN matrix.
        All entries must be an index in values
    values : Series
        pandas Series with index the nodes of the NxN matrix of which the simplices are listed,
        and values the values on that node to be averaged.
    with_multiplicity : bool
        if ``True`` the values are averaged with multiplicity i.e., they are weighted by the number of times a node
        participates in a simplex in a given position
        if ``False`` repetitions of a node in a given position are ignored.

    Returns
    -------
    DataFrame
        with index, the possible positions of o node in a ``k``-simplex and columns the mean, standard deviation and
        standard error of the mean for that position
    """
    # Filter values
    if with_multiplicity:
        vals_sl = values.loc[simplex_list.flatten()].to_numpy().reshape(simplex_list.shape)
    else:
        vals_sl = pd.concat([values.loc[np.unique(simplex_list[:, pos])] for pos in range(simplex_list.shape[1])],
                            axis=1, keys=range(simplex_list.shape[1]))
    # Compute stats
    stats_vals = pd.DataFrame(index=pd.Index(range(simplex_list.shape[1]), name="position"))
    # Stats per position
    stats_vals["mean"] = np.nanmean(vals_sl, axis=0)
    stats_vals["std"] = np.nanstd(vals_sl, axis=0)
    stats_vals["sem"] = stats.sem(vals_sl, axis=0, nan_policy="omit")
    # Stats in any position
    stats_vals.loc["all", "mean"] = np.nanmean(vals_sl)
    stats_vals.loc["all", "std"] = np.nanstd(vals_sl)
    stats_vals.loc["all", "sem"] = stats.sem(vals_sl, axis=None, nan_policy="omit")
    return stats_vals

def node_stats_per_position(simplex_lists, values, dims=None, with_multiplicity=True):
    """ Get across dimensions mean, standard deviation and standard error of the mean averaged across simplex lists
    and filtered per position
    Parameters
    ----------
    simplex_lists : dict
        keys : are int values representing dimensions
        values : for key ``k`` array of dimension (no. of simplices, ``k``) listing simplices to be considered.
        Each row corresponds to a list of nodes on a simplex indexed by the order of the nodes in an NxN matrix.
        All entries must be an index in values
    values : Series
        pandas Series with index the nodes of the NxN matrix of which the simplices are listed,
        and values the values on that node to be averaged.
    with_multiplicity : bool
        if ``True`` the values are averaged with multiplicity i.e., they are weighted by the number of times a node
        participates in a simplex in a given position
        if ``False`` repetitions of a node in a given position are ignored.
    dims : iterable
        dimensions for which to run the analysis, if ``None`` all the keys of simplex lists will be analyzed

    Returns
    -------
    dict
        keys the dimensions anlayzed and values for key ``k`` a DataFrame
        with index, the possible positions of o node in a ``k``-simplex and columns the mean, standard deviation and
        standard error of the mean for that position.
    """
    if dims is None:
        dims = simplex_lists.index
    stats_dict = {}
    for dim in tqdm(dims):
        sl = simplex_lists.loc[dim]
        stats_dict[dim] = node_stats_per_position_single(sl, values, with_multiplicity=with_multiplicity)
    return stats_dict


def node_stats_participation(participation, vals, condition=operator.eq, dims=None):
    """ Get statistics of the values in vals across nodes filtered using node participation
    Parameters
    ----------
    participation : DataFrame
        DataFrame of node participation with index the nodes in nodes of an NxN matrix to consider,
        columns are dimensions and values are node participation computed with
        [node_participation](network_topology.md#src.connalysis.network.topology.node_participation).
    values : Series
        pandas Series with index the nodes of the NxN matrix of where node participation has been computed
        and vals the values on that node to be averaged.
    condition : operator
        operator with which to filter the nodes. The default ``operator.eq`` filters nodes such that their maximal
        dimension of node participation is a given value.
        Alternatively, ``operator.ge`` filters nodes such that their maximal dimension of node participation is at least a given
        value.
    dims : iterable
        dimensions for which to run the analysis, if ``None`` all the columns of participation will be analyzed

    Returns
    -------
    DataFrame
        with index, the dimensions for which the analysis have been run and columns the statistics of the values in vals
        where the nodes have been grouped according to the condition given.

    See Also
    --------
    [node_stats_per_position_single](network_stats.md#src.connalysis.network.stats.node_stats_per_position_single):
    A similar function where the position of the nodes in the simplex are taken into account.  Note in particular that
    if condition = ``operator.ge`` the weighted_mean of this analyisis is equivalent than the value given by this function for position ``all``.
    However the computation using
    [node_participation](network_topology.md#src.connalysis.network.topology.node_participation)
    is more efficient.
    """
    par_df = participation.copy()
    if dims is None:
        dims = par_df.columns
    vals = vals.rename("values")
    par_df["max_dim"] = (par_df > 0).sum(axis=1) - 1  # maximal dimension a node is part of
    stats_vals = {}
    for dim in dims:
        mask = condition(par_df.max_dim, dim)
        c = pd.DataFrame(vals.loc[par_df[mask].index], columns=["values"])
        c["weight"] = par_df[mask][dim]
        w_mean = c.apply(np.product, axis=1).sum() / (c["weight"].sum())
        stats_vals[dim] = (c.shape[0],  # Number of nodes fulfilling the condition
                           np.nanmean(c["values"]),
                           np.nanstd(c["values"]),
                           stats.sem(c["values"], nan_policy="omit"),
                           w_mean  # mean weighted by participation
                           )
    stats_vals = pd.DataFrame.from_dict(stats_vals, orient="index",
                                        columns=["counts", "mean", "std", "sem", "weighted_mean"])
    stats_vals.index.name = "dim"
    return stats_vals


def node_stats_neighborhood(values, adj=None, pre=True, post=True, all_nodes=True, centers=None,
                            include_center=True, precomputed=False, neighborhoods=None):
    """ Get basic statistics of the property values on the neighbhood of the nodes in centers in the
    graph described by adj.
    Parameters
    ----------
    values : Series
        pandas Series with index the nodes of the NxN matrix of which the simplices are listed,
        and values the values on that node to be averaged.
    adj : sparse matrix or 2d array
        The adjacency matrix of the graph
    pre : bool
        If ``True`` compute the nodes mapping to the nodes in centers (the in-neighbors of the centers)
    post : bool
        If ``True`` compute the nodes that the centers map to (the out-neighbors of the centers)
    all_nodes : bool
        If ``True`` compute the neighbors of all nodes in adj, if ``False`` compute only the neighbors of the nodes
        listed in centers
    centers : 1d-array
        The indices of the nodes for which the neighbors need to be computed.  This entry is ignored if
        all_nodes is ``True`` and required if all_nodes is ``False``
    include_center : bool
        If ``True`` it includes the center in the computation otherwise it ignores it
    precomputed : bool
        If ``False`` it precomputes the neighbhorhoods in adj,
        if ``False`` it skips the computation and reads it fromt the input
    neighborhoods : DataFrame
        DataFrame of neighbhoord indices. Required if precomputed is ``True``

    Returns
    -------
    DataFrame
        with index, centers to be considered and columns the sum, mean, standard deviation and
        standard error of the mean of the values in that neighborhood.

    See Also
    --------
    [neighborhood_indices] (network_local.md#src.connalysis.network.local.neighborhood_indices):
    Function to precompute the neighborhood_indices that can be used if precomputed is set ``True``.
    Precomputing the neighborhoods would increase efficiency if multiple properties are averaged across neighborhoods.
    """

    # Single value functions for DataFrames
    def append_center(x):
        # To include center in the computation
        return np.append(x["center"], x["neighbors"])

    def mean_nbd(nbd_indices, v):
        df = v[nbd_indices]
        return [np.nansum(df), np.nanmean(df), np.nanstd(df), stats.sem(df, nan_policy="omit")]

    # Get neighborhoods
    if precomputed:
        assert isinstance(neighborhoods,
                          pd.Series), "If precomputed a Series of neighbhoords indexed by their center must be provided"
    else:
        assert (adj is not None), "If not precomputed and adjancecy matrix must be provided"
        neighborhoods = neighborhood_indices(adj, pre=pre, post=post, all_nodes=all_nodes, centers=centers)
    centers = neighborhoods.index
    if include_center:
        neighborhoods = neighborhoods.reset_index().apply(append_center, axis=1)
    else:
        neighborhoods = neighborhoods.reset_index(drop=True)
    stat_vals = pd.DataFrame.from_records(neighborhoods.map(lambda x: mean_nbd(x, values)),
                                          columns=["sum", "mean", "std", "sem"])
    stat_vals["center"] = centers
    return stat_vals.set_index("center")

def edge_stats_participation(participation, vals, condition=operator.eq, dims=None):
    """ Get statistics of the values in vals across edges filtered using edge participation

    Parameters
    ----------
    participation : DataFrame
        DataFrame of edge participation with index the edges of an NxN matrix to consider,
        columns are dimensions and values are edge participation.
    values : Series
        pandas Series with index the edges of the NxN matrix of which edge participation has been computed
        and vals the values on that edge to be averaged.
    condition : operator
        operator with which to filter the nodes. The default ``operator.eq`` filters nodes such that their maximal
        dimension of edge participation is a given value.
        Alternatively, ``operator.ge`` filters edges such that their maximal dimension of node participation is at least a given
        value.
    dims : iterable
        dimensions for which to run the analysis, if ``None`` all the columns of participation will be analyzed

    Returns
    -------
    DataFrame
        with index, the dimensions for which the analysis have been run and columns the statistics of the values in vals
        where the nodes have been grouped according to the condition given.
    """
    par_df = participation.copy()
    if dims is None:
        dims = par_df.columns
    vals = vals.rename("values")
    par_df["max_dim"] = (par_df > 0).sum(axis=1)  # maximal dimension an edge is part of. Note that edge participation in dimension 0 is 0
    stats_vals = {}
    for dim in dims:
        mask = condition(par_df.max_dim, dim)
        c = pd.DataFrame(vals.loc[par_df[mask].index], columns=["values"])
        c["weight"] = par_df[mask][dim]
        w_mean = c.apply(np.product, axis=1).sum() / (c["weight"].sum())
        stats_vals[dim] = (c.shape[0],  # Number of nodes fulfilling the condition
                           np.nanmean(c["values"]),
                           np.nanstd(c["values"]),
                           stats.sem(c["values"], nan_policy="omit"),
                           w_mean  # mean weighted by participation
                           )
    stats_vals = pd.DataFrame.from_dict(stats_vals, orient="index",
                                        columns=["counts", "mean", "std", "sem", "weighted_mean"])
    stats_vals.index.name = "dim"
    return stats_vals.drop(0)



'''###POTENTIALLY USEFUL OR DELETE

def edge_par_filter(edge_par_df, dims = None, id_mapping=None):
    # TODO: Get edge_par inside function and take adj instead?
    # Build df with index edges indexed by id_mapping and columns edge_participation per dimension and max_dimension
    # If id_mapping == None nodes are indexed by range(number of nodes) and edges accordingly
    # dims = None (do all dimensions), assert dims are in columns
    # out df
    pass
def node_par_filter(node_par_df, dims = None, id_mapping=None):
    # TODO: Get node_par inside function and take adj instead?
    # Build df with index nodex indexed by id_mapping and columns edge_participation per dimension and max_dimension
    # If id_mapping == None nodes are indexed by range(number of nodes)
    # dims = None (do all dimensions), assert dims are in columns
    # out df
    pass

def slist_edge_filter(slists, edge_type, dims=None, mapping_id=None)
    # TODO: Get slists inside function and take adj instead?
    # TODO: Multiple edge_types in one go?
    # Build df with index nodex indexed by id_mapping and columns edge_participation per dimension and max_dimension
    # If id_mapping == None nodes are indexed by range(number of nodes)
    # dims = None (do all dimensions) assert dims are in index
    # edge_type: function that takes dim and gives source an target
    # out dict
    pass

def slist_node_filter(slists, position, dims=None, mapping_id=None)
    # TODO: Get slists inside function and take adj instead?
    # TODO: Multiple positions in one go.
    # Build df with index nodex indexed by id_mapping and columns edge_participation per dimension and max_dimension
    # If id_mapping == None nodes are indexed by range(number of nodes)
    # dims = None (do all dimensions) assert dims are in index
    # position: position of node in simplex
    # out dict
    pass


#### STUFF FROM MICRONS PAPER

#Restriction to valid gids i.e., gids that are both in the connectome and that have recorded activity

def get_valid_gid(conn_gids, record_gids):
    # Return set of gids for which both included in the connectome and whose activity is measured
    ids, counts=np.unique(record_gids,return_counts=True)
    record_gids=ids[counts<=1]
    return np.intersect1d(conn_gids, record_gids)

def add_row_col_gid_and_mdim(df_par, all_gids, map_gids):
    # Add row and column gids to indexing to df of edge participation
    df=df_par.copy()
    df['max_dim']=((df>=1).sum(axis=1))
    df=df.set_index(keys=pd.MultiIndex.from_tuples(df.index)).reset_index(names=["row", "col"])
    index=pd.MultiIndex.from_tuples(zip(map_gids.loc[df.row].to_numpy(),
                                    map_gids.loc[df.col].to_numpy()))
    return df.set_index(index)

def get_valid_edges_mask(df, valid_gids):
    # Get mask of edge particiapation on edges between valid gids
    #df=df.reset_index(names=["row_gid", "col_gid"])
    df=df.reset_index()
    return np.logical_and(np.isin(df.level_0, valid_gids), np.isin(df.level_1, valid_gids))

# Format correlation matrix
def build_corr_df(corr, record_gids, valid_gids):
    # Store correlation as a Series with index the corresponding edge
    # Restrict correlation matrix to valid gids
    mask=np.isin(record_gids, valid_gids)
    corr=corr[np.ix_(mask, mask)]; N=corr.shape[0]
    # Get off diagonal entries
    id_up=np.triu_indices(N,k=1); id_down=np.tril_indices(N,k=-1)
    off_diag=(np.concatenate([id_up[0], id_down[0]]),
              np.concatenate([id_up[1], id_down[1]]))
    # Index by gids of entries
    index=pd.MultiIndex.from_tuples(zip(valid_gids[off_diag[0]],valid_gids[off_diag[1]]))
    return pd.Series(corr[off_diag], index=index)

# Analysis based on edge participation
def format_edge_list(s_list, source, target, map_gids):
    # Get list of edges in s_list from source to target indexed by their gids
    df=pd.DataFrame(s_list[:, [source,target]], columns=["row", "col"])
    index=pd.MultiIndex.from_tuples(zip(map_gids.loc[df.row].to_numpy(),
                                    map_gids.loc[df.col].to_numpy()))
    return df.set_index(index)

def stats_edge_par(edge_df, corr_df, condition):
    # Get stats for edges with ege participation according to condition
    # condition operator.eq (max dimension equals a given value)
    # condition operator.ge (max dimension is at least a certain value)
    dims=np.arange(1, edge_df.columns[-2]+1)
    stats={}
    for dim in dims:
        mask=condition (edge_df.max_dim, dim)
        c=pd.DataFrame(corr_df.loc[edge_df[mask].index], columns=["correlation"])
        c["weight"]=edge_df[mask][dim]
        w_mean=c.apply(np.product, axis=1).sum()/(c["weight"].sum())
        stats[dim]=(c.shape[0], c["correlation"].mean(), sem(c["correlation"]), w_mean)
    return pd.DataFrame.from_dict(stats, orient="index", columns=["counts", "mean", "sem", "weighted_mean"])

# Analysis based on simplex list
def last_edge(dim):
    return dim-1, dim

def stats_edge_position(simplex_lists, corr_df,map_gids,valid_gids, edge_selection=last_edge):
    # Get stats for edges in a simplex in the position given by edge_selection
    dims=simplex_lists.index[1:]
    stats={}
    for dim in tqdm(dims):
        source, target= edge_selection(dim)
        edges=format_edge_list(simplex_lists[dim], source,target,map_gids)
        edges=edges[get_valid_edges_mask(edges, valid_gids)]
        c=corr_df.loc[edges.index]
        stats[dim]=(c.shape[0], c.mean(), sem(c))
    return pd.DataFrame.from_dict(stats, orient="index", columns=["counts", "mean", "sem"])'''


# SPDX-FileCopyrightText: 2024 Blue Brain Project / EPFL
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Functions for network analysis on connectomes."""

from connalysis import (modelling,
                        network,
                        randomization)
__all__ = [
    "modelling","network","randomization", "rand_utils"
]


# SPDX-FileCopyrightText: 2024 Blue Brain Project / EPFL
#
# SPDX-License-Identifier: AGPL-3.0-or-later

""""Functions for getting parameters for modelling of random graphs."""

# TODO: Import all functions that the user might need
from .randomization import *
from .rand_utils import (adjust_bidirectional_connections)

__all__ = ["run_ER","run_SBM","run_DD2","run_DD3","run_DD2_block_pre", "configuration_model"]



# SPDX-FileCopyrightText: 2024 Blue Brain Project / EPFL
#
# SPDX-License-Identifier: AGPL-3.0-or-later

# Network analysis functions restricted to neighborhoods

# Author(s): D. Egas Santander, M. Reimann, JP. Smith, J. Lazovskis
# Last modified: 10/2023

import numpy as np
import scipy.sparse as sp
import pandas as pd


def neighborhood_indices(M, pre=True, post=True, all_nodes=True, centers=None):
    """Computes the indices of the neighbors of the nodes listed in centers

        Parameters
        ----------
        M : sparse matrix or 2d array
            The adjacency matrix of the graph
        pre : bool
            If ``True`` compute the nodes mapping to the nodes in centers (the in-neighbors of the centers)
        post : bool
            If ``True`` compute the nodes that the centers map to (the out-neighbors of the centers)
        all_nodes : bool
            If ``True`` compute the neighbors of all nodes in M, if ``False`` compute only the neighbors of the nodes
            listed in centers
        centers : 1d-array
            The indices of the nodes for which the neighbors need to be computed.  This entry is ignored if
            all_nodes is ``True`` and required if all_nodes is ``False``.

        Returns
        -------
        data frame
            indices: range from 0 to ``M.shape[0]`` if all_nodes is set to ``True``, otherwise centers.

            values: the neighbors of each center in the indices.

        Raises
        ------
        AssertionError
            If the matrix M is not square
        AssertionError
            If both pre and post are ``False``
        AssertionError
            If all_nodes is ``False`` but centers are not provided

    """
    assert M.shape[0] == M.shape[1], "The matrix is not square"
    assert np.logical_or(pre, post), "At least one of the pre/post parameters must be True"
    if all_nodes: centers = np.arange(M.shape[0])
    assert centers is not None, "If all_nodes == False and array of centers must be provided"

    M = M.tocoo() if sp.issparse(M) else sp.coo_matrix(M)

    base_df = pd.DataFrame({"row": M.row, "col": M.col})
    idxx = pd.Index(centers, name="center")
    nb_df = pd.Series([[]] * centers.shape[0], index=idxx, name="neighbors")
    # Restriction to requested centers
    if not all_nodes: res_df = base_df.apply(lambda _x: np.isin(_x, centers))

    if post:
        df = base_df[res_df['row']] if not all_nodes else base_df
        new_df = df.rename(columns={"row": "center", "col": "neighbors"}).groupby("center")["neighbors"].apply(list)
        nb_df = nb_df.combine(new_df, np.union1d, fill_value=[])
    if pre:
        df = base_df[res_df['col']] if not all_nodes else base_df
        new_df = df.rename(columns={"col": "center", "row": "neighbors"}).groupby("center")["neighbors"].apply(list)
        nb_df = nb_df.combine(new_df, np.union1d, fill_value=[])
    return nb_df.apply(lambda _x: _x.astype(int))


def submat_at_ind(M, ind):
    """Computes the submatrix of M on the nondes indexed by ind

    Parameters
    ----------
    M : matrix
        the adjacency matrix of the graph
    ind : 1d-array
        the indices on which to slice the matrix M

    Returns
    -------
    matrix
        the adjaceny matrix of the submatrix of M on the nodes in ind
    """
    if sp.issparse(M): M=M.tocsr()
    return  M[ind][:, ind]

def neighborhood(adj, v, pre=True, post=True, include_center=True, return_neighbors=False):
    """Gets the neighborhood of v in adj
            Parameters
            ----------
            adj : sparse matrix or 2d array
                The adjacency matrix of the graph
            pre : bool
                If ``True`` compute the submatrix on the nodes mapping to v (the in-neighbors of v)
            post : bool
                If ``True`` compute the submatrix on the nodes mapping to v (the out-neighbors of v)
            include_center : bool
                If ``True`` include v in the neighborhood
            return_neighbors : bool
                If ``True`` also return the indices in adj of the neighbors of v

            Returns
            -------
            matrix (sparse if adj is sparse)
                If pre = post = ``True`` it returns the full neighbohood of v

                If pre = ``True`` and post = ``False``it returns the in-neighborhood of v

                If pre = ``False`` and post = ``True``it returns the out-neighborhood of v

                If include_center = ``True``, then v is the first indexed node in the matrix,
                else it is excluded
        """
    nb_df=neighborhood_indices(adj, pre=pre, post=post, all_nodes=False, centers=np.array([v]))
    if sp.issparse(adj): adj=adj.tocsr()
    nb_ind = nb_df.loc[v]
    if include_center:
        nb_ind = np.append(v, nb_ind)
    if return_neighbors:
        return submat_at_ind(adj, nb_ind), nb_ind
    else:
        return submat_at_ind(adj, nb_ind)

def neighborhood_of_set_indices(M, node_set, pre=True, post=True):
    """Computes the indices of the neighbors of the nodes in node_set

            Parameters
            ----------
            M : sparse matrix or 2d array
                The adjacency matrix of the graph
            pre : bool
                If ``True`` compute the nodes mapping to the nodes in node_set (the in-neighbors of the node_set)
            post : bool
                If ``True`` compute the nodes that the centers map to node_set (the out-neighbors of the node_set)
            node_set : 1d-array
                The indices of the nodes for which the neighbors need to be computed.

            Returns
            -------
            array
                indices of the neighbhors of node_set

            Raises
            ------
            AssertionError
                If both pre and post are ``False``
    """
    assert np.logical_or(pre, post), "At least one of the pre/post parameters must be True"
    return np.unique(np.concatenate(neighborhood_indices(M,pre=pre,post=post,all_nodes=False,centers=node_set).values))

def neighborhood_of_set(M, node_set, pre=True, post=True, include_centers=True, return_neighbors=False):
    """Gets the neighborhood of the nodes in node_set
            Parameters
            ----------
            M : sparse matrix or 2d array
                The adjacency matrix of the graph
            node_set : array
                The indices of the nodes of which the neighborhood will be computed
            pre : bool
                If ``True`` compute the submatrix on the nodes mapping to node_set (the in-neighbors of node_set)
            post : bool
                If ``True`` compute the submatrix on the nodes that node_set maps to (the out-neighbors of node_set)
            include_center : bool
                If ``True`` include node_set in the graph
            return_neighbors : bool
                If ``True`` also return the indices in M of the neighbors of node_set

            Returns
            -------
            matrix (sparse if M is sparse)
                If pre = post = ``True`` it returns the full neighbohood of node_set

                If pre = ``True`` and post = ``False``it returns the in-neighborhood of node_set

                If pre = ``False`` and post = ``True``it returns the out-neighborhood of node_set
    """
    nodes=neighborhood_of_set_indices(M, node_set, pre, post)
    if include_centers:
        nodes=np.unique(np.concatenate([node_set, nodesA]))
    if isinstance(M, sp.coo_matrix): M=M.tocsr()
    # Slicing in different ways depending on matrix type
    if isinstance(M, np.ndarray): nbd= M[np.ix_(nodes, nodes)]
    if isinstance(M, sp.csr_matrix): nbd= M[nodes].tocsc()[:, nodes]
    if isinstance(M, sp.csc_matrix): nbd= M[:,nodes].tocsr()[nodes]
    if return_neighbors: return nbd, nodes
    else: return nbd

def property_at_neighborhoods(adj, func, pre=True, post=True,include_center=True,
                             all_nodes=True, centers=None, **kwargs):
    """Computes the property func on the neighborhoods of the centers within adj

            Parameters
            ----------
            adj : sparse matrix or 2d array
                The adjacency matrix of the graph
            func : function
                Function computing a network theoretic property e.g., degree or simplex counts
            pre : bool
                If ``True`` include the nodes mapping to the nodes in centers
            post : bool
                If ``True`` include the nodes that the centers map to
            include_center : bool
                If ``True`` include the centers
            all_nodes : bool
                If ``True`` compute func on the neighborhoods of all nodes in adj,

                If ``False`` compute it only the neighborhoods of the nodes listed in centers
            centers : 1d-array
                The indices of the nodes to consider.  This entry is ignored if
                all_nodes is ``True`` and required if all_nodes is ``False``.

            Returns
            -------
            dict
                keys: range from 0 to ``M.shape[0]`` if all_nodes is set to ``True``, otherwise centers

                values: the output of ``func`` in each neighborhood of the nodes in centers
    """
    nb_df=neighborhood_indices(adj, pre=pre, post=post, all_nodes=all_nodes, centers=centers)
    if sp.issparse(adj): adj=adj.tocsr()# TODO: Add warning of format!
    nbhd_values={}
    for center in nb_df.index:
        nb_ind = nb_df.loc[center]
        if include_center: nb_ind=np.append(center, nb_ind)
        nbhd=submat_at_ind(adj, nb_ind)
        #nbhd_values.loc[center]=func(nbhd, kwargs)
        nbhd_values[center]=func(nbhd, **kwargs)
    return nbhd_values


def properties_at_neighborhoods(adj, func_config, pre=True, post=True, include_center=True,
                                all_nodes=True, centers=None):
    """Computes the properties in func_config on the neighborhoods of the centers within adj

            Parameters
            ----------
            adj : sparse matrix or 2d array
                The adjacency matrix of the graph
            func_config : dict
                Configuration dictionary of functions to be computed on neihgborhoods
            pre : bool
                If ``True`` include the nodes mapping to the nodes in centers
            post : bool
                If ``True`` include the nodes that the centers map to
            include_center : bool
                If ``True`` include the centers
            all_nodes : bool
                If ``True`` compute func on the neighborhoods of all nodes in adj,

                If ``False`` compute it only the neighborhoods of the nodes listed in centers
            centers : 1d-array
                The indices of the nodes to consider.  This entry is ignored if
                all_nodes is ``True`` and required if all_nodes is ``False``.

            Returns
            -------
            dict
                keys: keys of func_config

                values: dict with

                    keys: range from 0 to ``M.shape[0]`` if all_nodes is set to ``True``, otherwise centers

                    values: the output of ``func`` in each neighborhood of the nodes in centers
    """
    nb_df = neighborhood_indices(adj, pre=pre, post=post, all_nodes=all_nodes, centers=centers)
    if sp.issparse(adj): adj = adj.tocsr()  # TODO: Add warning of format?
    nbhd_values = {key: {} for key in func_config.keys()}
    for center in nb_df.index:
        nb_ind = nb_df.loc[center]
        if include_center: nb_ind = np.append(center, nb_ind)
        nbhd = submat_at_ind(adj, nb_ind)
        for key in func_config.keys():
            nbhd_values[key][center] = func_config[key]['function'](nbhd, **func_config[key]['kwargs'])
    return nbhd_values

#### OLD CODE BELOW

def neighbours(v, matrix):
    #TODO: Delete redundant
    """Computes the neighbours of v in graph with adjacency matrix matrix

    Parameters
    ----------
    v : int
        the index of the vertex
    matrix : matrix
        the adjacency matrix of the graph

    Returns
    -------
    list
        the list of neighbours of v in matrix
    """
    neighbours = np.unique(np.concatenate((np.nonzero(matrix[v])[0],np.nonzero(np.transpose(matrix)[v])[0])))
    neighbours.sort(kind='mergesort')
    return np.concatenate((np.array([v]),neighbours))

def neighbourhood(v, matrix): #Previous name # def tribe(v, matrix):
    # TODO: Needs to be deleted
    """Computes the matrix induced by the neighbours of v in graph with adjacency matrix matrix

    Parameters
    ----------
    v : int
        the index of the vertex
    matrix : matrix
        the adjacency matrix of the graph

    Returns
    -------
    matrix
        the adjaceny matrix of the neighbourhood of v in matrix
    """
    #print(matrix.getformat())
    nhbd = neighbours(v, matrix)
    return matrix[np.ix_(nhbd,nhbd)]
#
# def local_property(adj, func, centers, kwargs):
#     nbhd_values=[]
#     for center in centers:
#         nbhd = neighbourhood(center, adj)
#         nbhd_values.append(func(nbhd, kwargs))
#     return nbhd_values

#TODO: Make this work without using the dataframe
# def top_chiefs(parameter, number=50, order_by_ascending=False):
# #  In: string, integer, boolean
# # Out: list of integers
#     return df.sort_values(by=[parameter],ascending=order_by_ascending)[:number].index.values

# def top_nbhds(parameter, number=50, order_by_ascending=False, matrix=adj):
# #  In: string, integer, boolean, matrix
# # Out: list of matrices
#     top_chief_list = top_chiefs(parameter, number=number, order_by_ascending=order_by_ascending)
#     return [neighbourhood(i, matrix=matrix) for i in top_chief_list]



#TODO: Not sure what this does, and it's not used. Should we include it?
# def new_nbhds(nbhd_list, index_range):
# #  In: list of list of integers
# # Out: list of list of integers
#     new_list = []
#     choice_vector = range(index_range)
#     for nbhd in nbhd_list:
#         new_neighbours = np.random.choice(choice_vector, size=len(nbhd)-1, replace=False)
#         while nbhd[0] in new_neighbours:
#             new_neighbours = np.random.choice(choice_vector, size=len(nbhd)-1, replace=False)
#         new_list.append(np.hstack((nbhd[0], new_neighbours)))
#     return new_list

